#!/usr/local/bin/php
<?php
/**
 * Flash Policy Service v0.9.c
 *
 * This script listens for <policy-file-request/> on port 843 and serves up
 * an xml crossdomain policy file. This sort of service is necessary if any
 * flash content is going to connect to sockets on the running host.
 *
 * I have made every effort to package everything you need into a single
 * file here, even though it could easily have been split into 3 or more
 * scripts.
 *
 * Requirements:
 *  - PHP 5 with sockets, pcntl, and posix extensions
 *  - Root access (to bind to a <1024 port)
 *
 * Use:
 *  Simply execute the script from the command line as root:
 *    # ./FlashPolicyService.php
 *  You can enable debug mode by invoking the script with '-d' as a parameter:
 *    # ./FlashPolicyService.php -d
 *  To stop the daemon, simply send it a SIGTERM and it should attempt to
 *  exit cleanly.
 *
 *  If you get 'bad interpreter' errors or the like, change the #! line to
 *  reflect the actual installed location of your php cli.
 *
 * Configuration:
 *  At present, there aren't very many config options. Simply edit the
 *  section immediately following this header. Options are commented.
 *
 * License:
 *  This code is made available under a Creative Commons Attribution 3.0
 *  License. Basically, you can use it however you like, but I would
 *  appreciate some credit when you do.
 *
 * Disclaimer:
 *  I make no guarantees that this code won't make your server explode in a
 *  shower of blue flames. However, I don't expect that it will. I am actually
 *  confident that this code will be helpful.
 *
 *  That said, this is still my beta release. If you find any bugs, please
 *  let me know so I can fix them.
 *
 * - Ammon Lauritzen [Apr 21, '08]
 *   http://ammonlauritzen.com/blog/2008/04/22/flash-policy-service-daemon/
 *
 * Changelog
 * 0.9.c
 *   - Fixed some typoes that were the result of how I combined everything
 *     into a single file. Thanks Alex!
 * 0.9.b
 *   - Original version to be posted at this url.
 * 0.9.a
 *   - Original proof of concept code posted online.
 */

/*** Config ***/

// where should we save the log output?
$log_filename = "/tmp/flash-policy.log";

// uncomment these lines to choose which user to run the daemon as
// default behavior is to look up and attempt to run as 'nobody'
# $daemon_uid = 99;
# $daemon_gid = 99;

// set this if you want to use an external file in stead of the default
$xml_filename = "";
$default_xml =
	'<'.'?xml version="1.0" encoding="UTF-8"?'.'>'.
    '<cross-domain-policy xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="http://www.adobe.com/xml/schemas/PolicyFileSocket.xsd">'.
        '<allow-access-from domain="*" to-ports="*" secure="false" />'.
        '<site-control permitted-cross-domain-policies="all" />'.
    '</cross-domain-policy>';

/*** You shouldn't have to edit anything below here ***/

/////////////////////////////////////////////////////////////////////////////
class Logger {
	public function __construct( $logfile ) {
        $this->logfile = $logfile;
        $this->open_log();
    }// end: constructor

    public function __destruct() {
        @fflush( $this->fh );
        @fclose( $this->fh );
    }// end: destructor

    public function log( $msg ) {
		// deal with redundant log spam
		if( $msg == $this->last_msg ) {
			$this->last_msg_count++;
			return;
		} else if( $this->last_msg_count ) {
			$this->write( "Last message repeated ".$this->last_msg_count." times." );
		}
		$this->lasg_msg = $msg;
		$this->last_msg_count = 0;

		// actually write the log message out
		$this->write( $msg );
	}// end: log

	private function write( $msg ) {
        $msg = sprintf("[%s] %s\n",date("y-m-d H:i:s"),$msg);
        $succ = @fwrite( $this->fh, $msg );
        if( $succ === FALSE ) {
            echo $msg;
        }
    }// end: write

	private function open_log() {
        if( !file_exists($this->logfile) ) {
            touch( $this->logfile );
            chmod( $this->logfile, 0664 );
        }
        $this->fh = @fopen( $this->logfile, "a" );
    }
}// end: logger class

/////////////////////////////////////////////////////////////////////////////
class Daemon {
	public function __construct() {
		error_reporting(0);
		set_time_limit(0);

		global $log_filename, $pid_filename;
		$this->logger = new Logger( $log_filename );
		$this->pid_filename = $pid_filename;

		$this->log_tag = $this->daemon_tag = "launcher";
		$this->debug( "constructed", $this->daemon_tag );
	}// end: constructor

	public function __destruct() {
		$this->debug( "destructing", $this->daemon_tag );
	}// end: destructor

	protected function log( $msg, $tag = false ) {
		if( $tag == false )
			$tag = $this->log_tag;
		$this->logger->log( $tag.": ".$msg );
	}// end: log
	protected function debug( $msg, $tag = false ) {
		if( DEBUG )
			$this->log( $msg, $tag );
	}// end: debug

	protected function main() {
		$this->log( "override main() in daemon subclass", $this->daemon_tag );
		$this->stop();
	}// end: main

	public function start() {
		$this->log( "starting daemon", $this->daemon_tag );
		if( !$this->_start() ) {
			$this->log( "unable to start daemon", $this->daemon_tag );
			return;
		}

		// report execution details
		$this->log( "uid = ".posix_getuid().", gid = ".posix_getgid() );
		$this->log( "cwd = ".getcwd() );

		// invoke main loop
		$this->running = true;
		while( $this->running ) {
			$this->main();
		}
	}// end: start
	private function _start() {
		if( !$this->_fork() ) {
			return false;
		}// end: try to fork

		if( !posix_setsid() ) {
			$this->log( "unable to setsid()", $this->daemon_tag );
			return false;
		}// end: try to set sid
		
		if( !$this->_suid() ) {
			return false;
		}// end: try to set uid
		
		// register signal handler
		declare(ticks = 1);
		pcntl_signal( SIGTERM, array(&$this, "on_sigterm") );
		// chdir somewhere moderately safe by default
		chdir( '/tmp' );
		return true;
	}// end: _start
	private function _fork() {
		$this->log( "forking...", $this->daemon_tag );
		$pid = pcntl_fork();
		if( $pid == -1 ) {
			// error
			$this->log( "unable to fork", $this->daemon_tag );
			return false;
		} else if( $pid ) {
			// parent
			$this->debug( "done with parent", $this->daemon_tag );
			exit( 0 );
		} else {
			// child
			$this->daemon_tag = "daemon";
			$this->child = true;
			$this->pid = posix_getpid();
			$this->debug( "child pid = ".$this->pid );
			return true;
		}
	}// end: _fork
	private function _suid() {
		global $daemon_uid, $daemon_gid;
		if( !isset($daemon_uid) || !isset($daemon_gid) ) {
			// we didn't get a uid/gid, try to make one up
			$this->debug( "searching for info on 'nobody'", $this->daemon_tag );
			$res = posix_getpwnam("nobody");
			if( $res !== FALSE ) {
				$uid = $res['uid'];
				$gid = $res['gid'];
			} else {
				// the 'nobody' user doesn't exist on this system, refuse
				// to daemonize as root
				$this->log( "unable to find info on 'nobody' user", $this->daemon_tag );
				return false;
			}
		} else {
			$this->debug( "got uid/gid of $daemon_uid/$daemon_gid", $this->daemon_tag );
			$uid = $daemon_uid;
			$gid = $daemon_gid;
		}
		// actually try to switch now
		if( !posix_setgid($gid) ) {
			$this->log( "unable to set gid to ".$gid, $this->daemon_tag );
			return false;
		} else if( !posix_setuid($uid) ) {
			$this->log( "unable to set uid to ".$uid, $this->daemon_tag );
			return false;
		} else {
			return true;
		}
	}// end: _suid

	public function stop() {
		$this->log( "stopping daemon", $this->daemon_tag );
		$this->running = false;
	}// end: stop

	protected function on_sigterm( $sig ) {
		if( $sig == SIGTERM ) {
			$this->log( "got SIGTERM", $this->daemon_tag );
			$this->stop();
			exit( 0 );
		}
	}// end: on_sigterm
}// end: daemon class

/////////////////////////////////////////////////////////////////////////////
class FlashPolicyService extends Daemon {
	public function __construct() {
		parent::__construct();
		$this->log_tag = "fps";
		$this->debug( "constructing" );

		$this->connections = array();
		$this->request_str = "<policy-file-request/>";
		$this->port = 843;

		// get our xml
		global $xml_filename, $default_xml;
		if( strlen($xml_filename) == 0 || !file_exists($xml_filename) ) {
			$this->log( "unable to read xml from '$xml_filename', using default" );
			$this->xml = $default_xml;
		} else {
			$this->log( "reading policy xml from $xml_filename" );
			$this->xml = file_get_contents( $xml_filename );
		}
		$this->debug( "policy xml: ".$this->xml );
		// make sure we're null terminated
		$this->xml = trim($this->xml)."\n\0";

		// and get going
		$this->init();
	}// end: constructor

	public function __destruct() {
		parent::__destruct();
		if( $this->sock )
			@fclose($this->sock);
	}// end: destructor

	private function init() {
		$this->debug( "init..." );
		if( $this->check_socket() ) {
			parent::start();
		} else {
			$this->log( "not starting daemon" );
		}
	}// end: init

	protected function main() {
		if( $this->sock ) {
			while( true ) {
				$this->accept_socket();
			}
		} else {
			$this->log( "server socket is closed?!" );
			parent::stop();
		}
		// paranoia to keep from absolutely hosing cpu if something goes wrong
		usleep( 10000 );	// 10ms
	}// end: main

	private function check_socket() {
		$this->sock = @socket_create( AF_INET, SOCK_STREAM, SOL_TCP );
		if( !$this->sock ) {
			$this->log( "unable to create socket" );
		} else {
			$succ = @socket_bind( $this->sock, "0.0.0.0", $this->port );
			if( !$succ ) {
				$this->log( "unable to bind to port ".$this->port );
			} else {
				$backlog = 100;
				$succ = @socket_listen( $this->sock, $backlog );
				if( !$succ ) {
					$this->log( "unable to listen with backlog of $backlog" );
				} else {
					// everything's good
					return true;
				}
			}// end: able to bind
		}// end: able to create socket

		// if we got here, it's an error. abort.
		$errno = socket_last_error( $this->sock );
		$errstr = socket_strerror( $errno );
		$this->log( "socket error: $errno: $errstr" );
		return false;
	}// end: check_socket

	private function accept_socket() {
		$r_socks = array_merge( array($this->sock), $this->connections );
		$this->debug( "selecting on ".count($r_socks)." sockets" );
		
		// block until something interesting happens
		$select = @socket_select( $r_socks, $w_socks = NULL, $e_socks = NULL, NULL );
		if( !$select )
			return;

		// did we get a new connection?
		if( in_array($this->sock, $r_socks) ) {
			$conn = @socket_accept( $this->sock );
			if( $conn !== false )
				@socket_getpeername( $conn, $addr );
			$this->log( "connection accepted from $addr, $conn" );
			$this->connections[] = $conn;
		}// end: got a new connection

		// check for policy requests
		foreach( $r_socks as $conn ) {
			// ignore the server socket
			if( $conn == $this->sock )
				continue;

			// read from the client
			$data = @socket_read( $conn, 1024 );
			if( $data === FALSE ) {
				$this->log( "got disconnect from $conn" );
			}// end: client closed connection
			else {
				$this->debug( "read '".trim($data)."' from $conn" );
				if( strpos($data, $this->request_str) !== FALSE ) {
					$this->log( "sending policy xml to $conn" );
					@socket_write( $conn, $this->xml );
				} else {
					$this->log( "got invalid request from $conn" );
				}
			}// end: got data from the client

			// and always disconnect after having read something, whether
			// it was a valid request or not - especially if it wasn't ;)
			@socket_close( $conn );
			$key = array_search( $conn, $this->connections );
			if( $key !== FALSE )
				unset( $this->connections[$key] );
		}// end: foreach socket
	}// end: accept_socket

}// end: flash policy service class

/////////////////////////////////////////////////////////////////////////////
/**
 * Actual execution code here. This checks if the php install has all of our
 * requisite extensions, makes sure we're launching as root, and checks if
 * debug mode was requested before actually starting the daemon.
 */

// make sure we have required extensions
$required_extensions = array( "sockets", "posix", "pcntl" );
$missing_extension = false;
foreach( $required_extensions as $ext ) {
	if( !extension_loaded($ext) ) {
		echo "Missing required php extension '$ext'.\n";
		$missing_extension = true;
	}
}
if( $missing_extension ) {
	exit( 1 );
}

// make sure we launch as root, otherwise we can't bind to 843
if( posix_getuid() != 0 || posix_getgid() != 0 ) {
	echo "Policy service must be started as root.\n";
	exit( 1 );
}

// see if we're in debug mode
define( "DEBUG", in_array('-d',$argv) );

// start the daemon
$fps = new FlashPolicyService();
?>
